# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import os

from betterboto import client as betterboto_client

from boto3 import client

import yaml
import click
import logging
import sys
from aws_organized import migrations
from aws_organized.extensions.service_control_policies import service_control_policies
from datetime import datetime

logging.disable(sys.maxsize)

STATE_FILE = "state.yaml"
SERVICE_CONTROL_POLICY = "SERVICE_CONTROL_POLICY"
ORGANIZATIONAL_UNIT = "ORGANIZATIONAL_UNIT"
META_FILE_NAME = "_meta.yaml"
SEP = os.path.sep
SSM_PARAMETER_PREFIX = "/_aws-organized/migrations"

EXTENSION = "aws_organized"


def list_policies_for_target(organizations: client, id: str, filter) -> dict:
    return organizations.list_policies_for_target(
        TargetId=id,
        Filter=filter,
    ).get("Policies", [])


def describe_organizational_unit(organizations: client, id: str) -> dict:
    return organizations.describe_organizational_unit(OrganizationalUnitId=id).get(
        "OrganizationalUnit"
    )


def list_children(organizations: client, id: str, child_type: str) -> dict:
    return organizations.list_children_single_page(
        ParentId=id,
        ChildType=child_type,
    ).get("Children", [])


def get_service_control_policies_for_target(organizations: client, id: str) -> dict:
    service_control_policies_for_target = dict()
    for policy in list_policies_for_target(organizations, id, SERVICE_CONTROL_POLICY):
        service_control_policies_for_target[policy.get("Id")] = policy
    return service_control_policies_for_target


def get_children_details_filter_by_organizational_unit(
    organizations: client, id: str, by_name: dict, parent_path: str, by_id: dict
) -> dict:
    result = dict()
    organizational_units = list_children(organizations, id, ORGANIZATIONAL_UNIT)
    for organizational_unit in organizational_units:
        organizational_unit_id = organizational_unit.get("Id")

        details = describe_organizational_unit(organizations, organizational_unit_id)
        path = parent_path + "/" + details.get("Name")
        by_name[path] = details.get("Id")
        by_id[details.get("Id")] = dict(path=path, details=details)

        result[organizational_unit_id] = dict(
            details=details,
            parent_id=id,
            policies=dict(
                service_control_policies=get_service_control_policies_for_target(
                    organizations, organizational_unit_id
                ),
            ),
            organizational_units=get_children_details_filter_by_organizational_unit(
                organizations, organizational_unit_id, by_name, path, by_id
            ),
        )
    return result


def update_state(role_arn) -> None:
    click.echo("Updating state file")
    with betterboto_client.CrossAccountClientContextManager(
        "organizations",
        role_arn,
        f"organizations",
    ) as organizations:
        all_accounts = dict()
        result = dict(accounts=all_accounts)
        list_roots_response = organizations.list_roots_single_page()
        tree = dict()
        by_name = dict()
        by_id = dict()
        organizational_units = dict(tree=tree, by_name=by_name, by_id=by_id)
        click.echo("Getting organizational units")
        result["organizational_units"] = organizational_units
        for root in list_roots_response.get("Roots", []):
            root_id = str(root.get("Id"))
            details = dict(
                Type="Root",
                Id=root_id,
                Name="Root",
            )
            by_name["/"] = root_id
            by_id[root_id] = dict(path="/", details=details)
            tree[root_id] = dict(
                details=details,
                organizational_units=get_children_details_filter_by_organizational_unit(
                    organizations, root_id, by_name, "", by_id
                ),
                policies=dict(
                    service_control_policies=organizations.list_policies_for_target(
                        TargetId=root_id,
                        Filter=SERVICE_CONTROL_POLICY,
                    ).get("Policies", []),
                ),
            )

        with open(STATE_FILE, "w") as f:
            f.write(yaml.safe_dump(result))
        click.echo("Saved organizational units")

        click.echo("Getting accounts")
        accounts = organizations.list_accounts_single_page().get("Accounts", [])
        click.echo(f"Adding accounts: {len(accounts)}")
        counter = 1
        for account in accounts:
            account_id = account.get("Id")
            click.echo(f"Adding {account_id} ({counter} of {len(accounts)})")
            all_accounts[account_id] = dict(
                details=account,
                parents=organizations.list_parents_single_page(ChildId=account_id).get(
                    "Parents"
                ),
                policies=dict(
                    service_control_policies=organizations.list_policies_for_target(
                        TargetId=account_id,
                        Filter=SERVICE_CONTROL_POLICY,
                    ).get("Policies", []),
                ),
            )
            counter += 1
        with open(STATE_FILE, "w") as f:
            f.write(yaml.safe_dump(result))
        click.echo("Saved accounts")

        click.echo("Finished")
        return result


def write_details(details: dict, output_dir: str) -> None:
    details_file = SEP.join([output_dir, META_FILE_NAME])
    with open(details_file, "w") as f:
        f.write(yaml.safe_dump(details))


def write_organizational_units(unit: dict, output_dir: str) -> None:
    details = unit.get("details")
    if details.get("Name") == "Root":
        name = details.get("Id")
    else:
        name = details.get("Name")

    this_output_dir = SEP.join([output_dir, name, "_organizational_units"])
    os.makedirs(this_output_dir, exist_ok=True)

    write_details(details, SEP.join([output_dir, name]))

    for child_unit_id, child_unit in unit.get("organizational_units", {}).items():
        write_organizational_units(
            child_unit,
            this_output_dir,
        )


def import_organization(role_arn: str, root_id: str) -> None:
    update_state(role_arn)
    state = yaml.safe_load(open(STATE_FILE, "r").read())
    output_dir = "environment"
    organizational_units = state.get("organizational_units").get("tree")
    by_id = state.get("organizational_units").get("by_id")

    root = organizational_units.get(root_id)
    write_organizational_units(
        root,
        output_dir,
    )

    # TODO: partition the state file by org id
    for account_id, account in state.get("accounts").items():
        account_details = account.get("details")
        parent_ou_id = account.get("parents")[0].get("Id")
        parent_ou_path = by_id.get(parent_ou_id).get("path")

        output_path_parts = [output_dir, root_id]
        for parent in parent_ou_path.split("/"):
            if parent != "":
                output_path_parts += [
                    "_organizational_units",
                    parent,
                ]
        output_path_parts += [
            "_accounts",
            account_details.get("Name"),
        ]
        output_path = SEP.join(output_path_parts)

        os.makedirs(output_path, exist_ok=True)
        with open(
            f"{output_path}{SEP}{META_FILE_NAME}",
            "w",
        ) as f:
            f.write(yaml.safe_dump(account_details))


def write_migration(
    extension: str, root_id: str, migration_type: str, migration_params: dict
) -> None:
    now = datetime.now()
    timestamp = datetime.timestamp(now)
    migration_file_name = f"{timestamp}_{migration_type}.yaml"
    os.makedirs(
        SEP.join(["environment", root_id, "_migrations"]),
        exist_ok=True,
    )
    with open(
        SEP.join(["environment", root_id, "_migrations", migration_file_name]), "w"
    ) as f:
        f.write(
            yaml.safe_dump(
                dict(
                    extension=extension,
                    migration_type=migration_type,
                    migration_params=migration_params,
                )
            )
        )


def make_migrations(role_arn: str, root_id: str) -> None:
    os.makedirs(SEP.join(["environment", "migrations"]), exist_ok=True)
    with betterboto_client.CrossAccountClientContextManager(
        "organizations",
        role_arn,
        f"organizations",
    ) as organizations:
        make_migrations_for_organizational_units(root_id, organizations)
        make_migrations_for_accounts(root_id, organizations)


def make_migrations_for_accounts(root_id: str, organizations) -> None:
    """
    Creates migrations for the following account use cases:
      - move an account
      - - when the remote parent ou exists - ACCOUNT_MOVE
      - - when the remote parent ou does not exist - ACCOUNT_MOVE_WITH_NON_EXISTENT_PARENT_OU
    :param organizations:
    :return:
    """
    accounts = get_accounts_folders()
    for account_file_path in accounts:
        account_name = account_file_path.split(SEP)[-1]
        account_details = yaml.safe_load(
            open(f"{account_file_path}{SEP}{META_FILE_NAME}", "r").read()
        )
        list_parents_single_page_response = organizations.list_parents_single_page(
            ChildId=account_details.get("Id")
        ).get("Parents")
        if len(list_parents_single_page_response) != 1:
            raise Exception(
                f"{account_details.get('Id')} has {len(list_parents_single_page_response)} parents."
            )
        remote_parent_organizational_unit_ou_id = list_parents_single_page_response[
            0
        ].get("Id")

        parent_ou_path_details_file_path = SEP.join(
            account_file_path.split(SEP)[0:-2] + [META_FILE_NAME]
        )

        if os.path.exists(parent_ou_path_details_file_path):
            local_parent_ou_details = yaml.safe_load(
                open(parent_ou_path_details_file_path, "r").read()
            )
            local_parent_organizational_unit_ou_id = local_parent_ou_details.get("Id")

            if (
                local_parent_organizational_unit_ou_id
                != remote_parent_organizational_unit_ou_id
            ):
                write_migration(
                    EXTENSION,
                    root_id,
                    migrations.ACCOUNT_MOVE,
                    dict(
                        account_id=account_details.get("Id"),
                        source_parent_id=remote_parent_organizational_unit_ou_id,
                        destination_parent_id=local_parent_organizational_unit_ou_id,
                    ),
                )
        else:
            destination_path = SEP.join(
                [""] + parent_ou_path_details_file_path.split(SEP)[3:-1]
            ).replace(f"{SEP}_organizational_units", "")
            write_migration(
                EXTENSION,
                root_id,
                migrations.ACCOUNT_MOVE_WITH_NON_EXISTENT_PARENT_OU,
                dict(
                    account_id=account_details.get("Id"),
                    source_parent_id=remote_parent_organizational_unit_ou_id,
                    destination_path=destination_path,
                ),
            )


def get_parent_ou_id_for_details_file(details_file_path: str) -> str:
    parent_path = SEP.join(details_file_path.split(SEP)[0:-2])
    if os.path.exists(f"{parent_path}{SEP}{META_FILE_NAME}"):
        parent_details = yaml.safe_load(
            open(f"{parent_path}{SEP}{META_FILE_NAME}", "r").read()
        )
        if parent_details.get("Id"):
            return parent_details.get("Id")
    return None


def make_migrations_for_organizational_units(organizations, root_id: str) -> None:
    """
    Creates migrations for the following OU use cases:
      - add an ou
      - - where the remote parent exists - OU_CREATE
      - - where the remote parent does not exist yet - OU_CREATE_WITH_NON_EXISTENT_PARENT_OU
      - rename an ou
      - - where the remote ou existed already - OU_RENAME

    Does not support the following OU use cases:
      - delete an ou
      - - where the remote ou exists already
      - - where the remote ou does not exist already
      - move an ou
      - - where there is already a remote target ou
      - - where there is not already a remote target ou
      - - where there is already a remote parent ou
      - - where there is not already a remote parent ou

    :param organizations:
    :return:
    """
    organizational_units_folders = get_organizational_units_folders()
    for organizational_unit_folder in organizational_units_folders:
        if os.path.exists(SEP.join([organizational_unit_folder, META_FILE_NAME])):
            details = yaml.safe_load(
                open(
                    SEP.join([organizational_unit_folder, META_FILE_NAME]),
                    "r",
                ).read()
            )
            remote_name = (
                organizations.describe_organizational_unit(
                    OrganizationalUnitId=details.get("Id")
                )
                .get("OrganizationalUnit")
                .get("Name")
            )

            local_name = organizational_unit_folder.split(SEP)[-1]
            if remote_name != local_name:
                write_migration(
                    EXTENSION,
                    root_id,
                    migrations.OU_RENAME,
                    dict(
                        name=local_name,
                        organizational_unit_id=details.get("Id"),
                    ),
                )

        else:
            parent_organizational_unit_folder = SEP.join(
                organizational_unit_folder.split(SEP)[0:-2]
            )
            new_ou_name = organizational_unit_folder.split(SEP)[-1]
            if os.path.exists(
                SEP.join([parent_organizational_unit_folder, META_FILE_NAME])
            ):
                parent_id = yaml.safe_load(
                    open(
                        SEP.join([parent_organizational_unit_folder, META_FILE_NAME]),
                        "r",
                    ).read()
                ).get("Id")
                write_migration(
                    EXTENSION,
                    root_id,
                    migrations.OU_CREATE,
                    dict(name=new_ou_name, parent_id=parent_id),
                )
            else:
                parent_ou_path = "/".join(
                    [""]
                    + parent_organizational_unit_folder.replace(
                        "_organizational_units", ""
                    )
                    .replace(f"{SEP}{SEP}", SEP)
                    .split(SEP)[2:]
                )
                write_migration(
                    EXTENSION,
                    root_id,
                    migrations.OU_CREATE_WITH_NON_EXISTENT_PARENT_OU,
                    dict(name=new_ou_name, parent_ou_path=parent_ou_path),
                )


def get_organizational_units_folders() -> list:
    return [
        x[0]
        for x in os.walk("environment/")
        if x[0].split(SEP)[-2] == "_organizational_units"
    ]


def get_accounts_folders() -> list:
    return [x[0] for x in os.walk("environment/") if x[0].split(SEP)[-2] == "_accounts"]


def migrate(role_arn: str) -> None:
    with betterboto_client.CrossAccountClientContextManager(
        "ssm",
        role_arn,
        f"ssm",
    ) as ssm:
        for migration_file in sorted(os.listdir("environment/migrations")):
            migration_id = migration_file.split(SEP)[-1].replace(".yaml", "")

            try:
                ssm.get_parameter(Name=f"{SSM_PARAMETER_PREFIX}/{migration_id}")
                click.echo(f"Migration: {migration_id} already run")
            except ssm.exceptions.ParameterNotFound:
                click.echo(
                    f"Record of migration: {migration_id} being run not found, running now"
                )
                migration = yaml.safe_load(
                    open(f"environment/migrations/{migration_file}", "r").read()
                )
                migration_extension = migration.get("extension")
                migration_type = migration.get("migration_type")
                migration_params = migration.get("migration_params")

                if migration_extension == EXTENSION:
                    migration_function = migrations.get_function(migration_type)
                elif migration_extension == service_control_policies.EXTENSION:
                    migration_function = (
                        service_control_policies.migrations.get_function(migration_type)
                    )

                try:

                    with betterboto_client.CrossAccountClientContextManager(
                        "organizations",
                        role_arn=role_arn,
                        role_session_name="ou_create",
                    ) as client:
                        result = migration_function(client, **migration_params)
                        status = "Ok" if result else "Failed"
                except Exception as ex:
                    status = "Unhandled error: {0}".format(ex)

                print(f"{migration_id}: { status }")
                ssm.put_parameter(
                    Name=f"{SSM_PARAMETER_PREFIX}/{migration_id}",
                    Description=f"Migration run: {datetime.utcnow()}",
                    Value=status,
                    Type="String",
                    Tags=[
                        {"Key": "AWS-Organized:Actor", "Value": "Framework"},
                    ],
                )



import glob
from datetime import datetime

import json
import yaml
import click
import os
from betterboto import client as betterboto_client
from . import migrations

SERVICE_CONTROL_POLICY = "SERVICE_CONTROL_POLICY"
ORGANIZATIONAL_UNIT = "ORGANIZATIONAL_UNIT"
ACCOUNT = "ACCOUNT"
SEP = os.path.sep

EXTENSION = "service_control_policies"


def save_all_policies_from_org(root_id: str, organizations) -> None:  # done maybe
    all_service_control_policies_in_org = organizations.list_policies_single_page(
        Filter=SERVICE_CONTROL_POLICY
    ).get("Policies")
    for policy in all_service_control_policies_in_org:
        described_policy = organizations.describe_policy(PolicyId=policy.get("Id")).get(
            "Policy"
        )
        policy_name = described_policy.get("PolicySummary").get("Name")
        path = f"environment/{root_id}/_policies/service_control_policies/{policy_name}"
        os.makedirs(path, exist_ok=True)
        open(
            f"{path}/policy.json",
            "w",
        ).write(json.dumps(json.loads(described_policy.get("Content")), indent=4))
        del described_policy["Content"]
        open(
            f"{path}/_meta.yaml",
            "w",
        ).write(yaml.safe_dump(described_policy.get("PolicySummary")))


def get_id_for_entity_from_path(entity_path: str) -> str:  # done
    return yaml.safe_load(open(entity_path, "r").read()).get("Id")


def write_policy_for_given_entity_path(  # done
    entity_path: str, application_method: str, policies: dict
) -> None:
    existing = dict()
    existing[application_method] = policies
    open(
        entity_path.replace("_meta.yaml", "_service_control_policies.yaml"), "w"
    ).write(yaml.safe_dump(existing))


def save_policies_for_entity(entity_path: str, organizations) -> None:  # done
    entity_path_parts = entity_path.split(SEP)
    entity_type = entity_path_parts[-3]
    click.echo(f"{entity_type} - {entity_path}")
    policies = organizations.list_policies_for_target_single_page(
        TargetId=get_id_for_entity_from_path(entity_path),
        Filter=SERVICE_CONTROL_POLICY,
    ).get("Policies")
    write_policy_for_given_entity_path(entity_path, "Attached", policies)


def save_policies_for_each_entity(entities: list, organizations) -> None:  # done
    for entity in entities:
        save_policies_for_entity(entity, organizations)


def save_targets_for_policy(root_id, organizations) -> None:  # done
    policies = glob.glob(
        f"environment/{root_id}/_policies/service_control_policies/*/*.yaml"
    )
    state = yaml.safe_load(open("state.yaml", "r").read())
    for policy_file in policies:
        click.echo("looking at")
        policy = yaml.safe_load(open(policy_file, "r").read())
        policy_id = policy.get("Id")
        targets = organizations.list_targets_for_policy_single_page(
            PolicyId=policy_id
        ).get("Targets", [])
        for target in targets:
            inherited = list()
            if target.get("Type") == ACCOUNT:
                account = (
                    state.get("accounts").get(target.get("TargetId")).get("details")
                )
                attached = glob.glob(
                    f"environment/{root_id}/**/{account.get('Name')}/_meta.yaml",
                    recursive=True,
                )
            elif target.get("Type") == ORGANIZATIONAL_UNIT:
                ou = (
                    state.get("organizational_units")
                    .get("by_id")
                    .get(target.get("TargetId"))
                    .get("details")
                )
                attached = glob.glob(
                    f"environment/{root_id}/**/{ou.get('Name')}/_meta.yaml",
                    recursive=True,
                )
                inherited += glob.glob(
                    f"environment/{root_id}/**/{ou.get('Name')}/_accounts/*/_meta.yaml",
                    recursive=True,
                )
                inherited += glob.glob(
                    f"environment/{root_id}/**/{ou.get('Name')}/_organizational_units/**/_meta.yaml",
                    recursive=True,
                )
            elif target.get("Type") == "ROOT":
                attached = glob.glob(
                    f"environment/{root_id}/_meta.yaml",
                    recursive=True,
                )
                inherited += glob.glob(
                    f"environment/{root_id}/_meta.yaml",
                    recursive=True,
                )

            else:
                raise Exception(f"Not handled type: {target.get('Type')}")

            if attached:
                assert len(attached) == 1
                attached = attached[0]
                output_path = attached.replace(
                    "_meta.yaml", "_service_control_policies.yaml"
                )
                output = dict(
                    Attached=list(),
                    Inherited=list(),
                )
                if os.path.exists(output_path):
                    output = yaml.safe_load(open(output_path, "r").read())
                output["Attached"].append(policy)
                open(output_path, "w").write(yaml.safe_dump(output))

                for thing in inherited:
                    output_path = thing.replace(
                        "_meta.yaml", "_service_control_policies.yaml"
                    )
                    output = dict(
                        Attached=list(),
                        Inherited=list(),
                    )
                    if os.path.exists(output_path):
                        output = yaml.safe_load(open(output_path, "r").read())
                    i = dict(Source=target.get("Name"))
                    i.update(policy)
                    output["Inherited"].append(i)
                    open(output_path, "w").write(yaml.safe_dump(output))


def remove_any_existing_policy_records(root_id: str) -> None:  # done
    policies = glob.glob(
        f"environment/{root_id}/**/_service_control_policies.yaml", recursive=True
    )
    for policy in policies:
        os.remove(policy)


def import_organization_policies(role_arn, root_id) -> None:  # done
    click.echo("Updating state file")
    with betterboto_client.CrossAccountClientContextManager(
        "organizations",
        role_arn,
        f"organizations",
    ) as organizations:
        remove_any_existing_policy_records(root_id)
        save_all_policies_from_org(root_id, organizations)
        save_targets_for_policy(root_id, organizations)


def check_policies(root_id: str, organizations) -> None:
    scps_path = (
        f"environment/{root_id}/_policies/service_control_policies/*/policy.json"
    )
    for policy_content_path in glob.glob(scps_path):
        policy_file_path = policy_content_path.replace("policy.json", "_meta.yaml")
        if os.path.exists(policy_file_path):
            local_policy = yaml.safe_load(open(policy_file_path, "r").read())
            p = organizations.describe_policy(PolicyId=local_policy.get("Id")).get(
                "Policy"
            )
            remote_policy = p.get("PolicySummary")
            if local_policy.get("Name") != remote_policy.get(
                "Name"
            ) or local_policy.get("Description") != remote_policy.get("Description"):
                write_migration(
                    EXTENSION,
                    root_id,
                    migrations.POLICY_DETAILS_UPDATE,
                    dict(
                        id=local_policy.get("Id"),
                        name=local_policy.get("Name"),
                        description=local_policy.get("Description"),
                    ),
                )
            local_policy_content = json.dumps(
                json.loads(open(policy_content_path, "r").read())
            )
            if local_policy_content != p.get("Content"):
                write_migration(
                    EXTENSION,
                    root_id,
                    migrations.POLICY_CONTENT_UPDATE,
                    dict(
                        id=local_policy.get("Id"),
                        content=local_policy_content,
                    ),
                )
        else:
            local_policy_content = json.dumps(
                json.loads(open(policy_content_path, "r").read())
            )
            write_migration(
                EXTENSION,
                root_id,
                migrations.POLICY_CREATE,
                dict(
                    name=policy_file_path.split(SEP)[-2],
                    content=local_policy_content,
                ),
            )


def write_migration(
    extension: str, root_id: str, migration_type: str, migration_params: dict
) -> None:
    now = datetime.now()
    timestamp = datetime.timestamp(now)
    migration_file_name = f"{timestamp}_{migration_type}.yaml"
    os.makedirs(
        SEP.join(["environment", root_id, "_migrations"]),
        exist_ok=True,
    )
    with open(
        SEP.join(["environment", root_id, "_migrations", migration_file_name]), "w"
    ) as f:
        f.write(
            yaml.safe_dump(
                dict(
                    extension=extension,
                    migration_type=migration_type,
                    migration_params=migration_params,
                )
            )
        )


def check_attachment(root_id: str, policy_file_path: str, organizations) -> None:
    local_policies = yaml.safe_load(open(policy_file_path, "r").read())
    meta = yaml.safe_load(
        open(
            policy_file_path.replace("_service_control_policies.yaml", "_meta.yaml"),
            "r",
        ).read()
    )
    remote_policies = organizations.list_policies_for_target_single_page(
        TargetId=meta.get("Id"), Filter=SERVICE_CONTROL_POLICY
    ).get("Policies")

    for local_policy in local_policies.get("Attached", []):
        found = False
        for remote_policy in remote_policies:
            if local_policy.get("Id") == remote_policy.get("Id"):
                found = True
        if not found:
            write_migration(
                EXTENSION,
                root_id,
                migrations.POLICY_ATTACH,
                dict(
                    policy_id=local_policy.get("Id"),
                    target_id=meta.get("Id"),
                ),
            )


def check_attachments(root_id: str, organizations) -> None:
    policies = glob.glob(
        f"environment/{root_id}/**/_service_control_policies.yaml", recursive=True
    )
    for policy_file_path in policies:
        check_attachment(root_id, policy_file_path, organizations)


def make_migrations(role_arn: str, root_id: str) -> None:
    with betterboto_client.CrossAccountClientContextManager(
        "organizations",
        role_arn,
        f"organizations",
    ) as organizations:
        check_policies(root_id, organizations)
        check_attachments(root_id, organizations)



# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import os

import click

from aws_organized import helpers
from aws_organized import aws_organized
from aws_organized.extensions.service_control_policies import service_control_policies


@click.group()
def cli():
    """cli"""
    pass


@cli.command()
@click.argument("role_arn")
def make_migration_policies(role_arn) -> None:
    service_control_policies.make_migration_policies(role_arn)


@cli.command()
@click.argument("role_arn")
def apply_migration_policies(role_arn) -> None:
    service_control_policies.apply_migration_policies(role_arn)


@cli.command()
@click.argument("role_arn")
def import_organization(role_arn):
    for root_id in os.listdir("environment"):
        if root_id in ["migrations", "Policies", "policies_migration"]:
            continue
        click.echo(f"Processing root_id: {root_id}")
        aws_organized.import_organization(role_arn, root_id)
        service_control_policies.import_organization_policies(role_arn, root_id)


@cli.command()
@click.option("--role-name", default="ImportOrganizationRole")
@click.option("--path", default="/AWSOrganized/")
@click.option("--assuming-resource", default="root")
@click.option("--output-format", default="yaml")
@click.argument("assuming-account-id")
def generate_import_organization_role_template(
    role_name: str,
    path: str,
    assuming_resource: str,
    output_format: str,
    assuming_account_id: str,
):
    click.echo(
        helpers.generate_import_organization_role_template(
            role_name,
            path,
            assuming_account_id,
            assuming_resource,
            output_format.lower(),
        )
    )


@cli.command()
@click.option("--role-name", default="ImportOrganizationRole")
@click.option("--path", default="/AWSOrganized/")
@click.option("--assuming-resource", default="root")
@click.option("--output-format", default="yaml")
@click.argument("assuming-account-id")
def provision_import_organization_role_stack(
    role_name: str,
    path: str,
    assuming_resource: str,
    output_format: str,
    assuming_account_id: str,
):
    helpers.provision_import_organization_role_stack(
        role_name,
        path,
        assuming_account_id,
        assuming_resource,
        output_format.lower(),
    )


@cli.command()
@click.argument("role_arn")
def make_migrations(role_arn):
    for root_id in os.listdir("environment"):
        if root_id in ["migrations", "Policies", "policies_migration"]:
            continue
        click.echo(f"Processing root_id: {root_id}")
        aws_organized.make_migrations(role_arn, root_id)
        service_control_policies.make_migrations(role_arn, root_id)


@cli.command()
@click.option("--role-name", default="MakeMigrationsRole")
@click.option("--path", default="/AWSOrganized/")
@click.option("--assuming-resource", default="root")
@click.option("--output-format", default="yaml")
@click.argument("assuming-account-id")
def generate_make_migrations_role_template(
    role_name: str,
    path: str,
    assuming_resource: str,
    output_format: str,
    assuming_account_id: str,
):
    click.echo(
        helpers.generate_make_migrations_role_template(
            role_name,
            path,
            assuming_account_id,
            assuming_resource,
            output_format.lower(),
        )
    )


@cli.command()
@click.option("--role-name", default="MakeMigrationsRole")
@click.option("--path", default="/AWSOrganized/")
@click.option("--assuming-resource", default="root")
@click.option("--output-format", default="yaml")
@click.argument("assuming-account-id")
def provision_make_migrations_role_stack(
    role_name: str,
    path: str,
    assuming_resource: str,
    output_format: str,
    assuming_account_id: str,
):
    helpers.provision_make_migrations_role_stack(
        role_name,
        path,
        assuming_account_id,
        assuming_resource,
        output_format.lower(),
    )


@cli.command()
@click.argument("role_arn")
def migrate(role_arn):
    aws_organized.migrate(role_arn)


@cli.command()
@click.option("--role-name", default="MigrateRole")
@click.option("--path", default="/AWSOrganized/")
@click.option("--assuming-resource", default="root")
@click.option("--output-format", default="yaml")
@click.argument("assuming-account-id")
def generate_migrate_role_template(
    role_name: str,
    path: str,
    assuming_resource: str,
    output_format: str,
    assuming_account_id: str,
):
    click.echo(
        helpers.generate_migrate_role_template(
            role_name,
            path,
            assuming_account_id,
            assuming_resource,
            output_format.lower(),
        )
    )


@cli.command()
@click.option("--role-name", default="MigrateRole")
@click.option("--path", default="/AWSOrganized/")
@click.option("--assuming-resource", default="root")
@click.option("--output-format", default="yaml")
@click.argument("assuming-account-id")
def provision_migrate_role_stack(
    role_name: str,
    path: str,
    assuming_resource: str,
    output_format: str,
    assuming_account_id: str,
):
    helpers.provision_migrate_role_stack(
        role_name,
        path,
        assuming_account_id,
        assuming_resource,
        output_format.lower(),
    )


@cli.command()
@click.option("--codepipeline-role-name", default="CodePipelineRole")
@click.option("--codepipeline-role-path", default="/AWSOrganized/")
@click.option("--codebuild-role-name", default="CodeBuildRole")
@click.option("--codebuild-role-path", default="/AWSOrganized/")
@click.option("--ssm-parameter-prefix", default="/-AWS-Organized")
@click.option("--output_format", default="yaml")
@click.argument("migrate-role-arn")
def generate_codepipeline_template(
    codepipeline_role_name: str,
    codepipeline_role_path: str,
    codebuild_role_name: str,
    codebuild_role_path: str,
    ssm_parameter_prefix: str,
    output_format: str,
    migrate_role_arn: str,
):
    click.echo(
        helpers.generate_codepipeline_template(
            codepipeline_role_name,
            codepipeline_role_path,
            codebuild_role_name,
            codebuild_role_path,
            ssm_parameter_prefix,
            output_format.lower(),
            migrate_role_arn,
        )
    )


@cli.command()
@click.option("--codepipeline-role-name", default="AWSOrganizedCodePipelineRole")
@click.option("--codepipeline-role-path", default="/AWSOrganized/")
@click.option("--codebuild-role-name", default="AWSOrganizedCodeBuildRole")
@click.option("--codebuild-role-path", default="/AWSOrganized/")
@click.option("--ssm-parameter-prefix", default="/-AWS-Organized")
@click.option("--output-format", default="yaml")
@click.argument("migrate-role-arn")
def provision_codepipeline_stack(
    codepipeline_role_name: str,
    codepipeline_role_path: str,
    codebuild_role_name: str,
    codebuild_role_path: str,
    ssm_parameter_prefix: str,
    output_format: str,
    migrate_role_arn: str,
):
    helpers.provision_codepipeline_stack(
        codepipeline_role_name,
        codepipeline_role_path,
        codebuild_role_name,
        codebuild_role_path,
        ssm_parameter_prefix,
        output_format.lower(),
        migrate_role_arn,
    )


if __name__ == "__main__":
    cli()



import troposphere
import yaml
from awacs import (
    organizations as awacs_organizations,
    aws,
    sts as awacs_sts,
)
import pkg_resources

from awacs.iam import ARN as IAM_ARN
from troposphere import iam, s3, codebuild, codecommit, codepipeline, ssm
from betterboto import client as betterboto_client


def generate_role_template(
    command: str,
    actions: list,
    role_name: str,
    path: str,
    assuming_account_id: str,
    assuming_resource: str,
    output_format: str,
) -> str:
    t = troposphere.Template()
    t.description = f"Role used to run the {command} command"

    t.add_resource(
        iam.Role(
            title="role",
            RoleName=role_name,
            Path=path,
            Policies=[
                iam.Policy(
                    PolicyName=f"{command}-permissions",
                    PolicyDocument=aws.PolicyDocument(
                        Version="2012-10-17",
                        Id=f"{command}-permissions",
                        Statement=[
                            aws.Statement(
                                Sid="1",
                                Effect=aws.Allow,
                                Action=actions,
                                Resource=["*"],
                            ),
                        ],
                    ),
                )
            ],
            AssumeRolePolicyDocument=aws.Policy(
                Version="2012-10-17",
                Id="AllowAssume",
                Statement=[
                    aws.Statement(
                        Sid="1",
                        Effect=aws.Allow,
                        Principal=aws.Principal(
                            "AWS", [IAM_ARN(assuming_resource, "", assuming_account_id)]
                        ),
                        Action=[awacs_sts.AssumeRole],
                    ),
                ],
            ),
        )
    )

    if output_format == "json":
        return t.to_json()
    else:
        return t.to_yaml()


def generate_import_organization_role_template(
    role_name: str,
    path: str,
    assuming_account_id: str,
    assuming_resource: str,
    output_format: str,
) -> str:
    return generate_role_template(
        "import-organizations",
        [
            awacs_organizations.ListRoots,
            awacs_organizations.ListPoliciesForTarget,
            awacs_organizations.ListAccounts,
            awacs_organizations.ListChildren,
            awacs_organizations.DescribeOrganizationalUnit,
            awacs_organizations.ListParents,
            awacs_organizations.ListPolicies,
            awacs_organizations.DescribePolicy,
            awacs_organizations.ListTargetsForPolicy,
        ],
        role_name,
        path,
        assuming_account_id,
        assuming_resource,
        output_format,
    )


def provision_stack(stack_name_suffix: str, template: str) -> None:
    with betterboto_client.ClientContextManager(
        "cloudformation",
    ) as cloudformation:
        cloudformation.create_or_update(
            StackName=f"AWSOrganized-{stack_name_suffix}",
            TemplateBody=template,
            Capabilities=["CAPABILITY_NAMED_IAM"],
        )


def provision_import_organization_role_stack(
    role_name: str,
    path: str,
    assuming_account_id: str,
    assuming_resource: str,
    output_format: str,
) -> None:
    template = generate_import_organization_role_template(
        role_name,
        path,
        assuming_account_id,
        assuming_resource,
        output_format,
    )
    provision_stack("import-organization-role", template)


def generate_make_migrations_role_template(
    role_name: str,
    path: str,
    assuming_account_id: str,
    assuming_resource: str,
    output_format: str,
) -> str:
    return generate_role_template(
        "make-migrations",
        [
            awacs_organizations.DescribeOrganizationalUnit,
            awacs_organizations.ListParents,
            awacs_organizations.DescribePolicy,
            awacs_organizations.ListPoliciesForTarget,
        ],
        role_name,
        path,
        assuming_account_id,
        assuming_resource,
        output_format,
    )


def provision_make_migrations_role_stack(
    role_name: str,
    path: str,
    assuming_account_id: str,
    assuming_resource: str,
    output_format: str,
) -> None:
    template = generate_make_migrations_role_template(
        role_name,
        path,
        assuming_account_id,
        assuming_resource,
        output_format,
    )
    provision_stack("make-migrations-role", template)


def generate_migrate_role_template(
    role_name: str,
    path: str,
    assuming_account_id: str,
    assuming_resource: str,
    output_format: str,
) -> str:
    return generate_role_template(
        "migrate",
        [
            awacs_organizations.CreateOrganizationalUnit,
            awacs_organizations.UpdateOrganizationalUnit,
            awacs_organizations.MoveAccount,
            awacs_organizations.CreatePolicy,
            awacs_organizations.UpdatePolicy,
            awacs_organizations.AttachPolicy,
        ],
        role_name,
        path,
        assuming_account_id,
        assuming_resource,
        output_format,
    )


def provision_migrate_role_stack(
    role_name: str,
    path: str,
    assuming_account_id: str,
    assuming_resource: str,
    output_format: str,
) -> None:
    template = generate_migrate_role_template(
        role_name,
        path,
        assuming_account_id,
        assuming_resource,
        output_format,
    )
    provision_stack("migrate-role", template)


def generate_codepipeline_template(
    codepipeline_role_name: str,
    codepipeline_role_path: str,
    codebuild_role_name: str,
    codebuild_role_path: str,
    ssm_parameter_prefix: str,
    output_format: str,
    migrate_role_arn: str,
) -> str:
    version = pkg_resources.get_distribution("aws-organized").version
    t = troposphere.Template()
    t.set_description(
        "CICD template that runs aws organized migrate for the given branch of the given repo"
    )

    project_name = "AWSOrganized-Migrate"
    repository_name = "AWS-Organized-environment"

    repo = t.add_resource(
        codecommit.Repository("Repository", RepositoryName=repository_name)
    )

    artifact_store = t.add_resource(
        s3.Bucket(
            "ArtifactStore",
            BucketEncryption=s3.BucketEncryption(
                ServerSideEncryptionConfiguration=[
                    s3.ServerSideEncryptionRule(
                        ServerSideEncryptionByDefault=s3.ServerSideEncryptionByDefault(
                            SSEAlgorithm="AES256"
                        )
                    )
                ]
            ),
        )
    )

    codepipeline_role = t.add_resource(
        iam.Role(
            "CodePipelineRole",
            RoleName=codepipeline_role_name,
            Path=codepipeline_role_path,
            ManagedPolicyArns=[
                "arn:aws:iam::aws:policy/AdministratorAccess",
            ],
            AssumeRolePolicyDocument=aws.PolicyDocument(
                Version="2012-10-17",
                Statement=[
                    aws.Statement(
                        Effect=aws.Allow,
                        Action=[awacs_sts.AssumeRole],
                        Principal=aws.Principal(
                            "Service", ["codepipeline.amazonaws.com"]
                        ),
                    ),
                ],
            ),
        )
    )

    codebuild_role = t.add_resource(
        iam.Role(
            "CodeBuildRole",
            RoleName=codebuild_role_name,
            Path=codebuild_role_path,
            ManagedPolicyArns=[
                "arn:aws:iam::aws:policy/AdministratorAccess",
            ],
            AssumeRolePolicyDocument=aws.PolicyDocument(
                Version="2012-10-17",
                Statement=[
                    aws.Statement(
                        Effect=aws.Allow,
                        Action=[awacs_sts.AssumeRole],
                        Principal=aws.Principal("Service", ["codebuild.amazonaws.com"]),
                    ),
                ],
            ),
        )
    )

    version_parameter = ssm.Parameter(
        "versionparameter",
        Name=f"{ssm_parameter_prefix}/version",
        Type="String",
        Value=version,
    )

    t.add_resource(version_parameter)

    project = t.add_resource(
        codebuild.Project(
            "AWSOrganizedMigrate",
            Artifacts=codebuild.Artifacts(Type="CODEPIPELINE"),
            Environment=codebuild.Environment(
                ComputeType="BUILD_GENERAL1_SMALL",
                Image="aws/codebuild/standard:4.0",
                Type="LINUX_CONTAINER",
                EnvironmentVariables=[
                    {
                        "Name": "MIGRATE_ROLE_ARN",
                        "Type": "PLAINTEXT",
                        "Value": migrate_role_arn,
                    },
                    {
                        "Name": "Version",
                        "Type": "PARAMETER_STORE",
                        "Value": troposphere.Ref(version_parameter),
                    },
                ],
            ),
            Name=project_name,
            ServiceRole=troposphere.GetAtt(codebuild_role, "Arn"),
            Source=codebuild.Source(
                Type="CODEPIPELINE",
                BuildSpec=yaml.safe_dump(
                    dict(
                        version="0.2",
                        phases=dict(
                            install={
                                "runtime-versions": dict(python="3.8"),
                                "commands": [
                                    "pip install aws-organized==${Version}",
                                ],
                            },
                            build={
                                "commands": [
                                    "aws-organized migrate $MIGRATE_ROLE_ARN",
                                ],
                            },
                        ),
                        artifacts=dict(
                            files=[
                                "environment",
                            ],
                        ),
                    )
                ),
            ),
        )
    )

    source_actions = codepipeline.Actions(
        Name="SourceAction",
        ActionTypeId=codepipeline.ActionTypeId(
            Category="Source",
            Owner="AWS",
            Version="1",
            Provider="CodeCommit",
        ),
        OutputArtifacts=[codepipeline.OutputArtifacts(Name="SourceOutput")],
        Configuration={
            "RepositoryName": repository_name,
            "BranchName": "main",
            "PollForSourceChanges": "true",
        },
        RunOrder="1",
    )

    t.add_resource(
        codepipeline.Pipeline(
            "Pipeline",
            RoleArn=troposphere.GetAtt(codepipeline_role, "Arn"),
            Stages=[
                codepipeline.Stages(
                    Name="Source",
                    Actions=[source_actions],
                ),
                codepipeline.Stages(
                    Name="Migrate",
                    Actions=[
                        codepipeline.Actions(
                            Name="Migrate",
                            InputArtifacts=[
                                codepipeline.InputArtifacts(Name="SourceOutput")
                            ],
                            ActionTypeId=codepipeline.ActionTypeId(
                                Category="Build",
                                Owner="AWS",
                                Version="1",
                                Provider="CodeBuild",
                            ),
                            Configuration={
                                "ProjectName": troposphere.Ref(project),
                                "PrimarySource": "SourceAction",
                            },
                            RunOrder="1",
                        )
                    ],
                ),
            ],
            ArtifactStore=codepipeline.ArtifactStore(
                Type="S3", Location=troposphere.Ref(artifact_store)
            ),
        )
    )

    if output_format == "json":
        return t.to_json()
    else:
        return t.to_yaml()


def provision_codepipeline_stack(
    codepipeline_role_name: str,
    codepipeline_role_path: str,
    codebuild_role_name: str,
    codebuild_role_path: str,
    ssm_parameter_prefix: str,
    output_format: str,
    migrate_role_arn: str,
) -> None:
    template = generate_codepipeline_template(
        codepipeline_role_name,
        codepipeline_role_path,
        codebuild_role_name,
        codebuild_role_path,
        ssm_parameter_prefix,
        output_format,
        migrate_role_arn,
    )
    provision_stack("codepipeline", template)
